;;;;;
;;;;; BRL - the Beautiful Report Language
;;;;;

;;;;; Part I: functions to manipulate generic scheme objects
;;;;; (brl-string brl-nonstring brl-symbol)

;;;
;;; (brl-string 4) ==> "4"
;;; (brl-nonstring "4") ==> 4
;;;

(define (brl-string x)
  (cond ((string? x) x)
	(else (let ((thunk (lambda () (display x))))
		(with-output-to-string thunk)))))

(define (brl-nonstring x)
  (cond ((string? x) (with-input-from-string x read))
	(else x)))

;;;
;;; brl-charset:symbol1
;;; contains all chars that are ok to use as first char in multi-char symbol
;;; brl-charset:symbol2
;;; contains all chars that are ok to use as second char in symbol
;;;

(define brl-charset:extended-alphabetic
  (chars->char-set (string->list "+-.*/<=>!?:$%_&~^")))

; The first char of an identifier may not include a number.
; For multi-char identifiers, + and - may not be 1st char.

(define brl-charset:symbol1
  (char-set-union char-set:alphabetic
		  (char-set-difference brl-charset:extended-alphabetic
				       (char-set #\+ #\-))))

(define brl-charset:symbol2
  (char-set-union char-set:alphanumeric brl-charset:extended-alphabetic))

(define (brl-symbol1? ch) (char-set-member? brl-charset:symbol1 ch))
(define (brl-symbol2? ch) (char-set-member? brl-charset:symbol2 ch))

;;;
;;; If a charlist won't safely make a symbol, return one that will:
;;; (brl-symbol-safe '(#\a #\space #\b) brl-symbol1?)
;;;    ==>  '(#\a #\_ #\b)
;;;

(define (brl-symbol-safe chl test)
  (cond ((null? chl) '())
	((test (car chl))		; legal char for this position
	 (cons (car chl) (brl-symbol-safe (cdr chl) brl-symbol2?)))
	(else (cons #\_ (brl-symbol-safe (cdr chl) brl-symbol2?)))))

;;;
;;; (brl-string->symbol "99 bottles of beer")
;;;   ==> '_9_bottles_of_beer
;;;

; In MIT Scheme, don't use string->symbol because of uppercase problems:
; (eq? 'PATH (string->symbol "PATH")) ==> #f
; (eq? 'PATH (intern "PATH"))         ==> #t

(define (brl-string->symbol str)
  (intern (list->string (brl-symbol-safe (string->list str) brl-symbol1?))))

(define (brl-symbol obj)
  (cond ((symbol? obj) obj)
	(else (brl-string->symbol (brl-string obj)))))

;;;;; Part II: functions to manipulate environments

;;;
;;; (brl-error-environment condition parent-environment)
;;;   ==> a child environment of paerent-environment
;;;       with useful bindings:
;;;         condition
;;;         report-string     from condition/report-string
;;;         field-names       from condition-type/field-names
;;;         [others]          one for each of field-names
;;;

(define (make-quote obj) (list 'quote obj))

(define (brl-error-environment condition parent-environment)
  (let* ((type (condition/type condition))
	 (field-names (condition-type/field-names type))
	 (field-value (lambda (field-name)
			((condition-accessor type field-name)
			 condition)))
	 (field-values (map field-value field-names))
	 (pair-list (map list field-names (map make-quote field-values))))
    (eval (list 'let (cons (list 'condition condition)
			   (cons (list 'report-string
				       (condition/report-string condition))
				 (cons (list 'error-fields
					     (make-quote field-names))
				       pair-list)))
		'(the-environment))
	  parent-environment)))

;;;
;;; Put a symbol/value binding into an environment if not already there
;;; If already in this env (not parents), append new value to list of values.
;;; This behavior is useful for HTML forms with multiple same-name inputs.
;;;

(define (brl-environment-put! env sym val)
  (cond ((memq sym (environment-bound-names env))
	 (let ((oldval (eval sym env)))
	   (cond ((list? oldval)
		  (environment-assign! env sym (append! oldval (list val))))
		 (else
		  (environment-assign! env sym (list oldval val))))))
	(else (eval (list 'define sym (list 'quote val)) env))))

(define (brl-environment-child parent-env)
  (eval '(let () (the-environment)) parent-env))

;;;
;;; (brl-environment-with-bindings '(a b) '(1 2))
;;;   ==> an environment with a bound to 1, b bound to 2
;;;

(define (brl-environment-with-bindings brl-symbols brl-values)
  (let () ; get a clean environment
    (let ((brl-local-environment (the-environment)))
      (let ((brl-put! (lambda (sym val) (brl-environment-put!
                                         brl-local-environment sym val))))
        (for-each brl-put! brl-symbols brl-values)
        brl-local-environment))))

(define (brl-make-bindings-env parent-env brl-symbols brl-values)
  (let ((brl-local-environment (eval '(let () (the-environment)) parent-env)))
    (let ((brl-put! (lambda (sym val) (brl-environment-put!
				       brl-local-environment sym val))))
      (for-each brl-put! brl-symbols brl-values)
      brl-local-environment)))

;;;;
;;;;  TABLES
;;;;

(define (brl-table-contents table) (assq 'contents table))
(define (brl-table-environments table) (assq 'environments table))
(define (brl-table-parent table) (assq 'parent table))
(define (brl-table? obj) (and (alist? obj)
			      (brl-table-contents obj)
			      (brl-table-environments obj)
			      (brl-table-parent obj)))

(define (brl-table-rows table)
  (if (null? table)
      '()
      (cdr (cadr (brl-table-contents table)))))

; Table contents are defined as a list of same-length lists

(define (brl-table-rowsize? obj rowsize)
  (cond ((null? obj) #t)
	((not (list? obj)) #f)
	((not (list? (car obj))) #f)
	((and (positive? rowsize)
	      (not (equal? (length (car obj)) rowsize))) #f)
	(else (brl-table-rowsize? (cdr obj) (length (car obj))))))
  
(define (brl-table-contents? obj)
  (brl-table-rowsize? obj -1))

(define (brl-make-table-env contents env)
  (cond ((brl-table-contents? contents)
	 (list (list 'contents contents)
	       (list 'environments '())
	       (list 'parent env)))
	(else (error "Invalid table" contents))))

(define (brl-make-table contents) (brl-make-table-env contents '()))

;;;
;;; (brl-environments-from-table-contents '((a b) (1 2) (3 4)))
;;;   ==> a list of two environments, one with a=1 and b=2
;;;       the other with a=3 and b=4
;;;

(define (brl-environments-from-table table)
  (let* ((contents (cadr (brl-table-contents table)))
	 (symbols (map brl-symbol (car contents)))
	 (parent (cadr (brl-table-parent table)))
	 (make-row (lambda (values) (brl-make-bindings-env
				     parent symbols values))))
    (map make-row (cdr contents))))

(define (brl-table-ensure-environments table)
  (let ((envs (brl-table-environments table)))
    (if (null? (cadr envs))
	(begin
	  (if (null? (cadr (brl-table-parent table)))
	      (set-cdr! (brl-table-parent table) (list (the-environment))))
	  (set-cdr! envs
		    (list (brl-environments-from-table table)))))
    table))

;;;
;;; (brl-bindings-env 'symbol 'value env)
;;;    ==> a list of environments, one for each binding in env
;;;        binding (a c) yields bindings ((symbol a) (value c))
;;;

(define (brl-bindings-env sym-for-sym sym-for-val env)
  (let ((make-env (lambda (binding) (brl-environment-with-bindings
				     (list sym-for-sym sym-for-val)
				     binding))))
    (map make-env (environment-bindings env))))

;;;;; Part III: functions to handle BRL report templates

(define brl-null-value "")

(define (brl-filter env)
  (let ((next-char (peek-char)))
    (cond ((eof-object? next-char)	; End of file
	   '())
	  ((char=? next-char #\[)	; Variable substitution, e.g. [email]
	   (begin
	     (read-char)		; skip the [
	     (let ((expr (read)))
	       (begin
		 (read-char)		; skip the ]
		 (let ((result (eval expr env)))
		   (if (null? result)
		       (display (eval 'brl-null-value env))
		       (display result)))))
	     (brl-filter env)))
	  (else				; Literal chars
	   (begin
	     (write-string (read-string (char-set #\[)))
	     (brl-filter env))))))

;;;
;;; (brl-filter-result env template-filename)
;;;  ==> either a string with the output results
;;;         or an error condition
;;;

(define (brl-filter-result env template-filename)
  (let* ((filter-thunk (lambda () (brl-filter env)))
	 (input-thunk (lambda () (with-input-from-file template-filename
				   filter-thunk)))
	 (string-thunk (lambda () (with-output-to-string input-thunk))))
    (ignore-errors string-thunk)))

;;;;; Part IV: functions to handle template repeats (e.g. tables)

;;;
;;; (brl-rest '(a b c d) '(a b)) ==> ((c d))
;;; (brl-rest '(a b) '(a b c))   ==> #f
;;;

(define (brl-rest l h)
  (cond ((null? h) (list l))		; list for when #f is '()
	((null? l) #f)
	((not (equal? (car l) (car h))) #f)
	(else (brl-rest (cdr l) (cdr h)))))

;;;
;;; Using MIT Scheme's read-string,
;;; efficiently get a reversed list of chars up to endchar
;;;

(define (brl-charsthrough endchar)
  (let ((s (read-string (char-set endchar))))
    (if (eof-object? s)
	'()
	(cons (read-char) (reverse (string->list s))))))

;;;
;;; with input from the string "the end is near"
;;; (brl-readto1 '() (reverse (string->list "end")))
;;;   ==> '(#\d #\n #\e #\Space #\e #\h #\t)
;;;   Leaves the stream pointing at " is near"
;;;

(define (brl-readto1 chars-read end-marker)
  (cond ((pair? chars-read)
	 (if (not (equal? (car chars-read) (car end-marker))) ; found EOF
	     (cdr chars-read)
	     (let ((r (brl-rest chars-read end-marker)))
	       (if r
		   chars-read
		   (let ((nextchars (brl-charsthrough (car end-marker))))
		     (if (null? nextchars)
			 chars-read
			 (brl-readto1
			  (append nextchars chars-read)
			  end-marker)))))))
	 (else
	  (let ((nextchars (brl-charsthrough (car end-marker))))
	    (if (null? nextchars)
		chars-read
		(brl-readto1
		 (append nextchars chars-read)
		 end-marker))))))

;;;
;;; with input from the string "the end is near"
;;; (brl-readthrough "end")
;;;   ==> "the end"
;;;   Leaves the stream pointing at " is near"
;;;

(define (brl-readthrough endstring)
  (list->string (reverse (brl-readto1 '() (reverse (string->list endstring))))))

;;;
;;; with input from the string "the end is near"
;;; (brl-readto "end")
;;;   ==> "the "
;;;   Leaves the stream pointing at " is near"
;;;

(define (brl-readto end-string)
  (let* ((end-marker (reverse (string->list end-string)))
	 (allch (brl-readto1 '() end-marker))
	 (rest (brl-rest allch end-marker)))
    (list->string
     (reverse
      (if (false? rest)
	  allch
	  rest)))))

(define (brl-repeat end-string table)
  (let* ((template-string (brl-readto end-string))
	 (thunk1 (lambda (env)
		   (let ((thunk2 (lambda () (brl-filter env))))
		     (with-input-from-string template-string thunk2)))))
    (for-each thunk1 (cadr (brl-table-environments
		      (brl-table-ensure-environments table)))))
  "")


;;;;;
;;;;; Misc convenience functions
;;;;;

(define (brl-separate-strings separator arglist)
  (if (null? arglist)
      '()
      (apply string-append (cons (car arglist)
				 (map (lambda (x) (string-append separator x))
				      (cdr arglist))))))

(define brl-separate
  (lambda args
    (case (length args)
      ((0 1) (error "brl-separate requires at least two arguments"))
      ((2)   (if (list? (cadr args))
		 (apply brl-separate-strings args)
		 (cadr args)))
      (else (brl-separate-strings (car args) (map brl-string (cdr args)))))))

(define (brl-nonblank? obj)
  (if (or (null? obj) (eof-object? obj))
      #f
      (string-find-next-char-in-set
       (brl-string obj) char-set:not-whitespace)))

; Remove blanks from list
(define (brl-nonblanks l)
  (if (null? l)
      l
      (if (brl-nonblank? (car l))
	  (cons (car l) (brl-nonblanks (cdr l)))
	  (brl-nonblanks (cdr l)))))

(define (brl-list obj)
  (cond ((null? obj) '())
	((brl-table? obj) (map car (brl-table-rows obj)))
	((list? obj) obj)
	(else (list obj))))

(define (brl-non-list obj)
  (if (pair? obj)
      (car obj)
      obj))
