;;;;;
;;;;; BRL-WWW: functions to integrate BRL with the World-Wide Web
;;;;;

;;;
;;; (brl-html-encode "foo->bar&baz") ==> "foo-&gt;bar&amp;baz"
;;;

(define brl-html-entity-amp (string->list "&amp;"))
(define brl-html-entity-gt (string->list "&gt;"))
(define brl-html-entity-lt (string->list "&lt;"))
(define brl-html-entity-quot (string->list "&quot;"))

(define (brl-html-encode-char ch)
  (case ch
    ((#\<) brl-html-entity-lt)
    ((#\>) brl-html-entity-gt)
    ((#\&) brl-html-entity-amp)
    ((#\") brl-html-entity-quot)
    (else (list ch))))

(define (brl-html-convert chl-from-reversed chl-to)
  (cond ((null? chl-from-reversed) chl-to)
	(else
	 (brl-html-convert
	  (cdr chl-from-reversed)
	  (append (brl-html-encode-char (car chl-from-reversed))
		  chl-to)))))

(define (brl-html-encode-string str)
  (list->string (brl-html-convert (reverse (string->list str)) '())))

(define (brl-html-encode x)
  (cond ((brl-table? x) (brl-html-encode-table x))
	(else (brl-html-encode-string (brl-string x)))))

(define (brl-html-encode-env x env)
  (cond ((null? x) (eval 'brl-null-value env))
	((brl-table? x) (brl-html-encode-table x))
	(else (brl-html-encode-string (brl-string x)))))

;;;
;;; (brl-url-decode "Hello+there%21")
;;;   ==>  "Hello there!"

(define (brl-url-decode-charlist chl)
  (cond ((null? chl) '())
	((eqv? (car chl) #\%)
	 (cons (integer->char
		(string->number
		 (list->string (list (cadr chl) (caddr chl)))
		 16))
	       (brl-url-decode-charlist (cdddr chl))))
	((eqv? (car chl) #\+)
	 (cons #\space (brl-url-decode-charlist (cdr chl))))
	(else (cons (car chl) (brl-url-decode-charlist (cdr chl))))))

(define (brl-url-decode str)
  (list->string (brl-url-decode-charlist (string->list str))))

;;;
;;; (brl-query-read "x=5&y=7&" env)
;;; has side effects of (define x "5") and (define y "7") within env
;;;

(define brl-semicolon (name->char ";"))

(define (brl-query-read-name input-name query-charlist env)
  (cond ((null? query-charlist) '())
	((eqv? (car query-charlist) #\=)
	 (brl-query-read-value input-name '() (cdr query-charlist) env))
	(else (brl-query-read-name
	       (append input-name (list (car query-charlist)))
	       (cdr query-charlist) env))))

(define (brl-query-read-value input-name input-value query-charlist env)
  (cond ((or (null? query-charlist)
	     (eqv? (car query-charlist) #\&)
	     (eqv? (car query-charlist) brl-semicolon))
	 ; We are at the end of a value in "name1=value1;name2=value2"
	 (begin
	   (brl-environment-put!
	    env
	    (intern
	     (list->string (brl-symbol-safe
			    (brl-url-decode-charlist input-name)
			    brl-symbol1?)))
	    (list->string (brl-url-decode-charlist input-value)))
	   (if (pair? query-charlist)
	       (brl-query-read-name '() (cdr query-charlist) env))))
	(else (brl-query-read-value input-name
				    (append input-value
					    (list (car query-charlist)))
				    (cdr query-charlist)
				    env))))

(define (brl-query-read query-string env)
  (brl-query-read-name '() (string->list query-string) env))

;;; (brl-silent (define a 1)) produces no output when used in a BRL template

(define brl-silent (lambda x ""))

(define (brl-cgi-env unix-vars query-string)
  (let ()
    (define brl-debug-time1 (get-decoded-time))
    ; put Unix environment variables in their own environment
    (brl-query-read unix-vars (the-environment))
    (define brl-debug-time2 (get-decoded-time))
    (let ()
      ; Put HTML form inputs in their own environment
      (brl-query-read query-string (the-environment))
      (define brl-debug-time3 (get-decoded-time))
      (the-environment))))

;
; Handle a CGI request with Unix environment variables,
; Form inputs (QUERY_STRING) and a template for output (PATH_TRANSLATED)
;
; (brl-cgi-request "" "email=brlewis@mit.edu;name=Bruce+Lewis;quest=grail%21;colour=green;" "test.txt")

; HTTP end-of-line marker (RFC 2068, section 2.2)
(define brl-http-eol
  (list->string (list (integer->char 13) (integer->char 10))))

(define (brl-cgi-start content-type)
  (string-append "Content-type: " content-type brl-http-eol brl-http-eol))

(define (brl-cgi-request unix-vars query-string template-filename)
  (let* ((env (brl-cgi-env unix-vars query-string))
	 (filter-result (brl-filter-result env template-filename))
	 (error-result
	  (cond ((condition? filter-result)
		 (brl-filter-result
		  (brl-error-environment filter-result env)
		  "htdocs/err.html"))
		(else filter-result)))
	 (final-result
	  (cond ((condition? error-result)
		 (string-append (condition/report-string filter-result)
				"<p>\n"
				(condition/report-string error-result)))
		(else error-result)))
	 (brl-buddy-separator (random 3/2)))
    (display "\nBRL-BUDDY: out\nBRL-SEPARATOR=")
    (display brl-buddy-separator)
    (display #\newline)
    (display (brl-cgi-start "text/html"))
    (display final-result)
    (display "BRL-SEPARATOR=")
    (display brl-buddy-separator)
    (display #\newline)))

(define (brl-lookup sym env)
  (if (symbol? sym)
      (eval sym env)
      (error "Bad symbol to look up: " sym)))

;;
;; Faster function for constructing scheme env from CGI request
;; (brl-cgi-env2 '((PATH "foo:bar") (HTTP_BAZ "quux"))
;;                '((x "5") (y "7")))
;; ==> an environment with bindings for inputs, whose parent has
;;     bindings for CGI environment variables
;;

(define (brl-cgi-env2 brl-cgi-bindings brl-input-bindings)
  (let ()
    (define brl-debug-time1 (get-decoded-time))
    ; put Unix environment variables in their own environment
    (define cgi-environment
	     (brl-make-bindings-env (the-environment)
				    (map car brl-cgi-bindings)
				    (map cadr brl-cgi-bindings)))
    (define brl-debug-time2 (get-decoded-time))
    ; Put HTML form inputs in their own environment
    (define input-environment
	     (brl-make-bindings-env cgi-environment
				    (map car brl-input-bindings)
				    (map cadr brl-input-bindings)))
    input-environment))

(define (brl-cgi-req2 cgi-bindings input-bindings template-filename)
  (let* ((env (brl-cgi-env2 cgi-bindings input-bindings))
	 (filter-result (brl-filter-result env template-filename))
	 (error-result
	  (cond ((condition? filter-result)
		 (brl-filter-result
		  (brl-error-environment filter-result env)
		  "htdocs/err.html"))
		(else filter-result)))
	 (final-result
	  (cond ((condition? error-result)
		 (string-append (condition/report-string filter-result)
				"<p>\n"
				(condition/report-string error-result)))
		(else error-result)))
	 (brl-buddy-separator (random 3/2)))
    (display "\nBRL-BUDDY: out\nBRL-SEPARATOR=")
    (display brl-buddy-separator)
    (display #\newline)
    (display (brl-cgi-start "text/html"))
    (display final-result)
    (display "BRL-SEPARATOR=")
    (display brl-buddy-separator)
    (display #\newline)))

;;;
;;; functions for rendering objects as HTML
;;;
(define (brl-html-obj obj null-val)
  (if (null? obj) null-val obj))

(define (brl-html-cell obj)
  (display "<TD>")
  (display obj)
  (display "</TD>\n"))

(define (brl-html-header obj)
  (display "<TH>")
  (display obj)
  (display "</TH>\n"))

(define (brl-html-row objlist)
  (display "<TR>\n")
  (for-each brl-html-cell objlist)
  (display "</TR>\n"))

(define (brl-html-table-contents attribs contents null-val)
  (let* ((html-obj (lambda (obj) (brl-html-obj obj null-val)))
	 (html-list (lambda (l) (map html-obj l))))
    (display (string-append "<TABLE " attribs ">\n"))
    (display "<TR>\n")
    (for-each brl-html-header (html-list (car contents)))
    (display "</TR>\n")
    (for-each brl-html-row (map html-list (cdr contents)))
    (display "</TABLE>\n")
    ""))

(define (brl-html-encode-table table)
  (if (zero? (length (brl-table-rows table)))
      ""
      (let* ((table-parent (cadr (brl-table-parent table)))
	     (parent (if (null? table-parent)
			 (the-environment)
			 table-parent))
	     (thunk (lambda () (brl-html-table-contents
				"BORDER=2 CELLPADDING=2"
				(cadr (brl-table-contents table))
				(eval 'brl-null-value parent)))))
	(with-output-to-string thunk))))

;;;
;;; (brl-html-close-tag "table border=2>")
;;;   ==> "</table>"
;;;

(define (brl-html-close-tag open-tag)
  (let ((hasspace (substring? " " open-tag)))
    (string-append "</" (if hasspace
			    (string-append (substring open-tag 0 hasspace) ">")
			    open-tag))))

;;;
;;; (brl-html-read) returns a list of 3 items, e.g.
;;; ("<ul>" "\n<li> [foo]\n" "</ul>")
;;; as read from HTML input
;;;

(define (brl-html-read-start)
  ; Read all junk up to a start tag
  (let* ((leader (brl-readthrough "<"))
	 (html-tag (brl-readthrough ">")))
    ; See if it's an end tag
    (if (equal? "/" (substring html-tag 0 1))
	(let ((next-start (brl-html-read-start)))
	  (cons (string-append leader html-tag (car next-start))
		(cdr next-start)))
	; else we've found the start tag
	(list leader html-tag))))

(define (brl-html-read)
  (let* ((start (brl-html-read-start))
	 (close-tag (brl-html-close-tag (cadr start))))
    (if (string-ci=? close-tag "</table>")
	; reading an HTML table which should just containt two rows
	(let* ((header-row-start (brl-html-read-start))
	       (header-row-end (brl-readthrough
				(brl-html-close-tag
				 (cadr header-row-start))))
	       (contents-row-start (brl-html-read-start))
	       (contents-row-end (brl-readthrough
				  (brl-html-close-tag
				   (cadr contents-row-start))))
	       (table-end (brl-readthrough
			   (brl-html-close-tag (cadr start)))))
	  (list
	   ; Print the table start and header row once
	   (apply string-append (append start header-row-start
					(list header-row-end)))
	   ; Then repeat the contents row
	   (apply string-append (append contents-row-start
					(list contents-row-end)))
	   ; Then the end
	   table-end))
	; else this is not a table
	(let* ((body-end (brl-readthrough close-tag))
	       (body (substring body-end 0 (- (string-length body-end)
					      (string-length close-tag)))))
	  (list (apply string-append start)
		body
		close-tag)))))

(define (brl-html-repeat-env table env)
  (let ((html-template (brl-html-read))
	(thunk1 (lambda () (brl-filter env))))
    (begin
      (with-input-from-string (car html-template) thunk1)
      (let ((thunk2 (lambda () (brl-repeat (caddr html-template) table))))
	(with-input-from-string (cadr html-template) thunk2))
      (with-input-from-string (caddr html-template) thunk1)))
  "")

(define (brl-html-mail-env env)
  (let* ((html-template (brl-html-read))
	 (thunk1 (lambda () (brl-filter env)))
	 (thunk2 (lambda () (with-input-from-string (cadr html-template)
			      thunk1)))
	 (filter-result (with-output-to-string thunk2)))
    (string-append (car html-template)
		   (brl-html-encode-env
		    (brl-mail-string (string-trim-left filter-result))
		    env)
		   (caddr html-template))))

(define (brl-html-select name default options)
  (let* ((option-list (brl-list options))
	 (with-default (if (member default option-list)
			   option-list
			   (cons default option-list)))
	 (make-option (lambda (str)
			(string-append
			 "  <OPTION"
			 (if (equal? str default)
			     " SELECTED>"
			     ">")
			 (brl-string str) "\n"))))
    (string-append
     "<SELECT NAME=\"" name "\">\n"
     (apply string-append (map make-option with-default))
     "</SELECT>\n")))

;;;
;;; Crude linkification functions
;;; assumes URLs appear in whitespace, not angle brackets, etc.
;;;

(define (brl-linkify-substring str start end)
  (string-append (substring str 0 start)
		 "<a href=\""
		 (substring str start end)
		 "\">"
		 (substring str start end)
		 "</a>"
		 (substring str end (string-length str))))

(define (brl-subsubstring? pattern str start)
  (let ((pos (substring? pattern (string-tail str start))))
    (if pos
	(+ pos start)
	#f)))

; This is an alternate char-set for ending urls
(define brl-charset:url-end
  (char-set-union char-set:whitespace (char-set #\> #\" #\&)))

; Takes a string with url's delimited by whitespace and linkifies the url's.

(define (brl-linkify-string str start)
  (let ((urlh (brl-subsubstring? "http://" str start))
	(urlf (brl-subsubstring? "ftp://" str start))
	(urlg (brl-subsubstring? "gopher://" str start))
	(len (string-length str)))
    (let ((urlstart (min (if (false? urlh) len urlh)
			 (if (false? urlf) len urlf)
			 (if (false? urlg) len urlg))))
      (if (equal? len urlstart)
	  str
	  (let ((urlend (substring-find-next-char-in-set
			 str urlstart len char-set:whitespace)))
	    (display urlstart)
	    (newline)
	    (display urlend)
	    (newline)
	    (brl-linkify-string
	     (brl-linkify-substring str urlstart urlend)
	     (+ urlstart 15 (* 2 (- urlend urlstart))))))))) ; end of URL
